import datetime
import gc
import inspect
import time

import pytest
import torch
import torch.nn as nn
from torch.backends import cudnn

from heavyball import PaLMForeachSOAP, SFPaLMForeachSOAP, PaLMForeachSFAdamW, PrecondScheduleSFPaLMSOAP, ForeachADOPT, \
    ForeachSOAP, ForeachSFAdamW, ForeachLaProp, PrecondScheduleForeachSOAP, PrecondSchedulePaLMForeachSOAP

steps = 10
cudnn.benchmark = True
cudnn.deterministic = False
torch.use_deterministic_algorithms(False)
torch.set_float32_matmul_precision("high")  # highest: FP32, high: TF32, medium: bf16

args = {'betas': (0.9, 0.95), 'precondition_frequency': 2, 'merge_dims': False, 'warmup_steps': 100,
        'max_precond_dim': 2 ** 16, 'beta': 0.9}


@pytest.mark.parametrize('opt', [PaLMForeachSOAP, SFPaLMForeachSOAP, PaLMForeachSFAdamW, PrecondScheduleSFPaLMSOAP,
                                 ForeachADOPT, ForeachSOAP, ForeachSFAdamW, ForeachLaProp, PrecondScheduleForeachSOAP,
                                 PrecondSchedulePaLMForeachSOAP, torch.optim.AdamW, torch.optim.Adam])
@pytest.mark.parametrize('dtype', [torch.float32, torch.bfloat16, torch.double])
@pytest.mark.parametrize('size,batch', [(128, 128)])
@pytest.mark.parametrize('lr', [0.1, 1e-2, 1e-3, 1e-4])
@pytest.mark.parametrize('weight_decay', [1e-2])
def test_f(opt, dtype, size, batch, lr, weight_decay):
    torch.cuda.empty_cache()
    gc.collect()
    time.sleep(1)
    torch.manual_seed(0x1239121)

    a = nn.Linear(size, size, bias=False).cuda().to(dtype)
    signature = inspect.signature(opt)
    o = opt(a.parameters(), lr, weight_decay=weight_decay,
            **{k: v for k, v in args.items() if k in signature.parameters})
    torch.cuda.empty_cache()
    gc.collect()

    loss_mean = 0
    start = datetime.datetime.now()

    for i in range(steps):
        inp = torch.randn((batch, size), device='cuda', dtype=dtype)
        loss = (a(inp) - inp).square().mean()
        loss.backward()
        o.step()
        o.zero_grad()
        with torch.no_grad():
            loss_mean += loss.detach() / steps

    if hasattr(o, 'eval'):
        o.eval()
    eval_loss = (a(inp) - inp).square().mean()
    dist = datetime.datetime.now() - start
    print(f'Took {dist} | {opt.__name__} | Loss: {loss.item()} - Eval: {eval_loss.item()} - Mean: {loss_mean.item()}')



from .foreach_adopt import ForeachADOPT
from .foreach_laprop import ForeachLaProp
from .foreach_sfadamw import ForeachSFAdamW
from .foreach_soap import ForeachSOAP
from .palm_foreach_sfadamw import PaLMForeachSFAdamW
from .palm_foreach_soap import PaLMForeachSOAP
from .precond_schedule_foreach_soap import PrecondScheduleForeachSOAP
from .precond_schedule_palm_foreach_soap import PrecondSchedulePaLMForeachSOAP
from .precond_schedule_sfpsoap import PrecondScheduleSFPaLMSOAP
from .schedule_free_palm_foreach_soap import SFPaLMForeachSOAP

PalmForEachSoap = PaLMForeachSOAP

__all__ = ['PalmForEachSoap', 'PaLMForeachSFAdamW', 'PaLMForeachSOAP', 'SFPaLMForeachSOAP', 'PrecondScheduleSFPaLMSOAP',
           'ForeachSOAP', 'ForeachSFAdamW', 'ForeachLaProp', 'ForeachADOPT', 'PrecondScheduleForeachSOAP',
           'PrecondSchedulePaLMForeachSOAP']



import torch
import torch.optim

from .utils import schedule_free_, warmup, ScheduleFree, exp_avg_sq_, beta_debias


class PaLMForeachSFAdamW(ScheduleFree):
    def __init__(self, params, lr=0.0025, beta=0.9, betas=(None, None), eps=1e-8, weight_decay=0, warmup_steps=0, r=0.0,
                 weight_lr_power=2.0, beta2_scale: float = 0.8):
        if betas[0] is not None:
            beta = betas[0]
        defaults = dict(lr=lr, beta=beta, eps=eps, r=r, k=0, warmup_steps=warmup_steps, train_mode=True,
                        weight_sum=0.0, lr_max=-1.0, weight_lr_power=weight_lr_power, weight_decay=weight_decay,
                        beta2_scale=beta2_scale)
        super().__init__(params, defaults)

    def step(self, closure=None):
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """

        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            eps = group['eps']
            decay = group['weight_decay']
            k = group['k']

            if not group['train_mode']:
                raise Exception("Not in train mode!")

            active_p = [p for p in group['params'] if p.grad is not None]

            for p in active_p:
                if 'z' not in self.state[p]:
                    self.state[p]['z'] = torch.clone(p.data)
                    self.state[p]['exp_avg_sq'] = torch.zeros_like(p.data, dtype=torch.float32)

            y, grad, exp_avg_sq, z = zip(
                *[(p.data, p.grad.float(), self.state[p]['exp_avg_sq'], self.state[p]['z']) for p in active_p])

            # Decay the first moment running average coefficient
            beta2 = 1 - (k + 1) ** -group['beta2_scale']
            old_debiased = beta_debias(beta2, k + 1)

            # Decay the first and second moment running average coefficient
            denom = exp_avg_sq_(exp_avg_sq, grad, old_debiased, eps)

            # Normalize grad in-place for memory efficiency
            torch._foreach_div_(grad, denom)

            # Weight decay calculated at y
            if decay != 0:
                torch._foreach_add_(grad, y, alpha=decay)

            lr = warmup(group['lr'], k + 1, group['warmup_steps'])
            group['weight_sum'] = schedule_free_(lr, group['weight_lr_power'], group['weight_sum'], group['beta'],
                                                 y, z, grad)

            group['k'] = k + 1
        return loss



import torch
import torch.optim

from .utils import warmup, exp_avg_sq_, beta_debias, update_param_


class ForeachLaProp(torch.optim.Optimizer):

    def __init__(self, params, lr=0.0025, betas=(0.9, 0.99), eps=1e-8, weight_decay=0, warmup_steps=0):
        defaults = dict(lr=lr, betas=betas, eps=eps, k=0, warmup_steps=warmup_steps, train_mode=True, weight_sum=0.0,
                        lr_max=-1.0, weight_decay=weight_decay)
        super().__init__(params, defaults)

    def step(self, closure=None):
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """

        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            eps = group['eps']
            decay = group['weight_decay']
            k = group['k']

            if not group['train_mode']:
                raise Exception("Not in train mode!")

            active_p = [p for p in group['params'] if p.grad is not None]

            for p in active_p:
                if 'exp_avg' not in self.state[p]:
                    self.state[p]['exp_avg'] = torch.zeros_like(p.data, dtype=torch.float32)
                    self.state[p]['exp_avg_sq'] = torch.zeros_like(p.data, dtype=torch.float32)

            y, grad, exp_avg_sq, exp_avg = zip(
                *[(p.data, p.grad.float(), self.state[p]['exp_avg_sq'], self.state[p]['exp_avg']) for p in active_p])

            # Decay the first and second moment running average coefficient
            denom = exp_avg_sq_(exp_avg_sq, grad, beta_debias(group['betas'][1], k + 1), eps)
            beta1 = beta_debias(group['betas'][0], k + 1)
            torch._foreach_mul_(exp_avg, beta1)
            torch._foreach_addcdiv_(exp_avg, grad, denom, 1 - beta1)
            del grad

            # Normalize grad in-place for memory efficiency
            lr = -warmup(group['lr'], k + 1, group['warmup_steps'])
            update_param_(y, exp_avg, lr, decay)

            group['k'] = k + 1
        return loss
